%{
/****************************************************************/
/* Lexical analyser specification for Integration Framework     */
/* configuration control file                                   */
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "variables.h"
#include "parser.h"

/* This redefines the input mechanism */
/* for flex so that we can read in    */
/* entire lines and handle errors in  */
/* a better way                       */

#define BUFFER_LENGTH 256
static char  kwd_line[BUFFER_LENGTH];
static char  last_kwd_line[BUFFER_LENGTH];
static int   kwd_eof;
static int   kwd_primed;
static int   kwd_pos;
static char *kwd_ptr;
static char *kwd_pathname;
extern char *expectedString;
int yylineno;

/* These bits are required if using GNU flex */
static int read_from_line(char *buf, int max);
#undef YY_INPUT
#define YY_INPUT(buf, nchars, max) (nchars = read_from_line(buf, max))

/* These bits are required if using ATT lex */
static char get_kchar(void);
static void unget_kchar(char c);

#undef input
#define input() get_kchar()
#undef unput
#define unput(c) unget_kchar(c)

%}

%option noyywrap
%o 3000
%%

[Dd][Ee][Ff][Aa][Uu][Ll][Tt] { return SetYYLVAL(DT_DEFAULT); }

[Ss][Ll][Ee][Ee][Pp]"_"[Tt][Ii][Mm][Ee] {
			  return SetYYLVAL(DT_SLEEPTIME);
			}

[Ss][Oo][Uu][Rr][Cc][Ee]"_"[Rr][Ee][Ff]  {
                          return SetYYLVAL(DT_SOURCEREF);
                        }
[Qq][Mm][Aa][Nn][Aa][Gg][Ee][Rr]  {
                          return SetYYLVAL(DT_QMANAGER);
                        }

[Qq][Nn][Aa][Mm][Ee]    { return SetYYLVAL(DT_QNAME); }

[Rr][Ee][Pp][Ll][Yy]"_"[Qq][Nn][Aa][Mm][Ee] {
                          return SetYYLVAL(DT_RQNAME);
                        }
[Ee][Rr][Rr][Oo][Rr]"_"[Qq][Nn][Aa][Mm][Ee] {
                          return SetYYLVAL(DT_ERRQNAME);
                        }
[Ee][Rr][Rr][Oo][Rr]"_"[Ff][Ii][Ll][Ee][Nn][Aa][Mm][Ee] {
                          return SetYYLVAL(DT_ERRFNAME);
                        }
[Ee][Rr][Rr][Oo][Rr]"_"[Mm][Aa][Ii][Ll]"_"[Aa][Dd][Dd][Rr] {
                          return SetYYLVAL(DT_MAILADDR);
                        }
[Mm][Ee][Ss][Ss][Aa][Gg][Ee]"_"[Ss][Ii][Zz][Ee] {
                          return SetYYLVAL(DT_QMSGSIZE);
                        }
[Ww][Aa][Rr][Nn][Ii][Nn][Gg][Ss]"_"[Aa][Ll][Ll][Oo][Ww][Ee][Dd] {
                          return SetYYLVAL(DT_WARNINGS);
                        }
[Ss][Ff][Mm]"_"[Ff][Ll][Aa][Vv][Oo][Uu][Rr] {
                          return SetYYLVAL(DT_SFMFLV);
                        }
[Ss][Ff][Mm]"_"[Ll][Aa][Ss][Tt][Ii][Dd][Ff][Ii][Ll][Ee] {
                          return SetYYLVAL(DT_SFMIDFILE);
                        }
[Aa][Cc][Kk]"_"[Ii][Dd][Ff][Ii][Ll][Ee] {
                          return SetYYLVAL(DT_ACKIDFILE);
                        }
[Ff][Ii][Ll][Ee]"_"[Dd][Ii][Rr][Ee][Cc][Tt][Oo][Rr][Yy] {
                          return SetYYLVAL(DT_FILEDIR);
                        }
[Ff][Ii][Ll][Ee]"_"[Ll][Ee][Aa][Ff][Nn][Aa][Mm][Ee] {
                          return SetYYLVAL(DT_FILELEAF);
                        }
[Ii][Nn][Tt][Ee][Gg][Ee][Rr] {
			  return SetYYLVAL(DT_INTEGER_VAR);
			}
[[Ff][Ll][Oo][Aa][Tt] {
			  return SetYYLVAL(DT_FLOAT_VAR);
			}
[Ss][Tt][Rr][Ii][Nn][Gg] {
			  return SetYYLVAL(DT_STRING_VAR);
			}
[Dd][Aa][Tt][Aa][Bb][Aa][Ss][Ee]"_"[Ll][Oo][Gg][Ii][Nn] {
			  return SetYYLVAL(DT_DBLOGIN);
			}
\"[^"\n]+\"             { return SetYYLVAL(DT_STRING); }
[a-zA-Z][a-zA-Z0-9$_]+  { return SetYYLVAL(DT_VARNAME); }
[0-9]+"."[0-9]+         { return SetYYLVAL(DT_FLOAT); }
[0-9]+                  { return SetYYLVAL(DT_INTEGER); }
\"\"                    { return SetYYLVAL(DT_NULLSTRING); }
"="                     { return SetYYLVAL(DT_EQUALS); }
\n                      { return SetYYLVAL(DT_NEWLINE); }
[ \t]+                  { kwd_pos += yyleng; }
^#.*$                   { ; }
.                       { kwd_pos++; return yytext[0]; }

%%

/******************************************************************/
/* Set the value of the lexicon returned to YACC                  */
/******************************************************************/

int SetYYLVAL(int Token)
{
#ifdef DEBUG
  printf("TOKEN: %.*s\n", yyleng, yytext);
#endif

/* Increment the character position */
/* by the length of the lexicon     */
kwd_pos += yyleng;

/* Perform any conversion required */
#ifdef YYCHAR_ARRAY
switch (Token)
  {
  case DT_INTEGER: yylval.longInt  = atol(yytext); break;
  case DT_FLOAT  : yylval.floatNum = atof(yytext); break;
  default        : yylval.string   = yytext;       break;
  }
#else
switch (Token)
  {
  case DT_INTEGER: yylval.longInt  = atol((char *)yytext); break;
  case DT_FLOAT  : yylval.floatNum = atof((char *)yytext); break;
  default        : yylval.string   = (char *)yytext;       break;
  }
#endif

return Token;
}

/******************************************************************/
/* Print an error message from the parser                         */
/******************************************************************/

int yyerror(char *string)
{
int lineLen = strlen(kwd_line) - 1;

fprintf(stderr, "*** ERROR *** [%s] in config file [%s] at line %d\n",
        string, kwd_pathname, yylineno, kwd_pos - yyleng);
fprintf(stderr, "*** ERROR *** expected %s\n", expectedString);
fprintf(stderr, "*** ERROR *** offending line follows\n--->\t%.*s\t<---\n",
        lineLen, kwd_line);
fprintf(stderr, "--->\t%*s^%*s\t<---\n", kwd_pos - yyleng, "",
        (lineLen - 1) - (kwd_pos - yyleng), "");
return 1;
}

/******************************************************************/
/* Read in the configuration control file                         */
/******************************************************************/

int read_config_file(char *configFileName,
                     int   max)
{
char *cptr;
int   len;

/* Test the input pathname for spaces. */
/* When the first one is found it is   */
/* turned into a NUL character. This   */
/* is done so that things like MQ that */
/* cannot NUL terminate parameters can */
/* call this function.                 */

for (cptr = configFileName, len = 0; len < max; cptr++, len++)
  {
  if (*cptr == '\0') break;
  if (*cptr == ' ')
    {
    *cptr = '\0';
    break;
    }
  }

/* If could not find one, IE the   */
/* pathname is as long as possible */
/* then need to copy the string to */
/* give enough space for the NUL   */

if (len == max)
  {
  if ((cptr = malloc(max + 1)) == NULL) return 1;
  strncpy(cptr, configFileName, max);
  cptr[max] = '\0';
  configFileName = cptr;
  }
 else
  cptr = NULL;

/* Open the given file for reading */
if ((yyin = fopen(kwd_pathname = configFileName, "r")) == NULL)
  {
  free(cptr);
  return 1;
  }

/* Call the parser on the input file */
if (yyparse())
  {
  free(cptr);
  return 1;
  }

/* Close the file afterwards */
fclose(yyin);

free(cptr);
return 0;
}

/*********************************************************************/
/* Read an entire line from the file                                 */
/*********************************************************************/

static int get_kline(void)
{
/* Backup the last line */
strcpy(last_kwd_line, kwd_line);

if (fgets(kwd_line, BUFFER_LENGTH, yyin) == NULL)
  {
#ifdef DEBUG
  fprintf(stderr, "EOF found");
#endif
  kwd_eof = 1;
  return 1;
  }

#ifdef DEBUG
  fprintf(stderr, "Line read = \"%s\"\n", kwd_line);
#endif

/* Set the position indicator to */
/* the start of the line         */

kwd_pos = 0;
kwd_ptr = kwd_line;

/* A new line was read in */
yylineno++;

#ifdef DEBUG
  printf("KLINE[%04d]:%s", yylineno, kwd_line);
#endif

return 0;
}

/*********************************************************************/
/* Read a character from the file line buffer                        */
/*********************************************************************/

static char get_kchar(void)
{
/* If end-of-file return */
/* end-of-string char    */

if (kwd_eof == 1) return 0;

/* If this the first time */
/* prime the input line   */

if (kwd_primed == 0)
  {
  if (get_kline()) return 0;
  kwd_primed = 1;
  }

/* Keep reading characters from the input */
/* file until a valid character is found  */

while (1)
  {
#ifdef DEBUG
  printf("KCHAR:%c\n", *kwd_ptr);
#endif

  switch (*kwd_ptr)
    {
    case '\0' : /* This was the end of the input line */
                /* so need to get another one         */

                if (get_kline()) return 0;
                continue;

    default   : /* Found a good character  */
                /* so return it            */

                return *kwd_ptr++;
    }
  }
}

/*********************************************************************/
/* Return a character to the file line buffer                        */
/*********************************************************************/

static void unget_kchar(char c)
{
if (kwd_primed == 0) return;
if (kwd_eof == 1) return;

#ifdef DEBUG
  printf("UKCHAR:%c\n", c);
#endif

if ((kwd_ptr - 1) != kwd_line) kwd_ptr--;
}

/****************************************************************/
/* Read in at most max characters from the input file           */
/****************************************************************/

static int read_from_line(char *buf,
                          int   max)
{
int len;

#ifdef DEBUG
  printf("Requested %d chars\n", max);
#endif

for (len = 0; len < max; len++)
  {
  if ((*buf++ = get_kchar()) == 0) break;
  
  /* Return each time hit newline */
  if (buf[-1] == '\n') return len + 1;
  }

return len;
}
